from typing import Any, Dict

from pydantic_compat import BaseModel

try:
    from pydantic import model_serializer
except ImportError:
    model_serializer = None  # type: ignore


class KindMixin(BaseModel):
    """Mixin to a `kind` field to the dict output.

    This helps for casting a dict to a specific subclass, when the fields are
    otherwise identical.
    """

    def __init__(self, **data: Any) -> None:
        data.pop("kind", None)
        return super().__init__(**data)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        d = super().dict(**kwargs)
        d["kind"] = self.__class__.__name__.lower()
        return d

    if model_serializer is not None:

        @model_serializer(mode="wrap")
        def serialize_root(self, handler, _info) -> Dict:  # type: ignore
            d = handler(self)
            d["kind"] = self.__class__.__name__.lower()
            return d



"""Vendorized dependencies.

Color has been moved from pydantic to pydantic-extra-types.  However, we can't easily
require pydantic-extra-types because it pins pydantic to >2.0.0.  So, in order to
retain compatibility with pydantic 1.x and 2.x without getting the color warning, and
without accidentally pinning pydantic, we vendor the two Color classes here.
"""

import pydantic.version

if pydantic.version.VERSION.startswith("2"):
    from ._pydantic_color_v2 import Color  # noqa
else:
    from ._pydantic_color_v1 import Color  # type: ignore # noqa

__all__ = ["Color"]



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Callable

from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.dataclass.generator import DataclassGenerator
from xsdata.utils import text
from xsdata.utils.collections import unique_sequence

from xsdata_pydantic_basemodel.pydantic_compat import PYDANTIC2

if TYPE_CHECKING:
    from xsdata.codegen.models import Attr, Class
    from xsdata.models.config import GeneratorConfig, OutputFormat

text.stop_words.update(("schema", "validate"))


class PydanticBaseGenerator(DataclassGenerator):
    """Python pydantic dataclasses code generator."""

    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        return PydanticBaseFilters(config)


class PydanticBaseFilters(Filters):
    def __init__(self, config: GeneratorConfig):
        super().__init__(config)
        self.pydantic_support = getattr(config.output, "pydantic_support", False)
        if self.pydantic_support == "both":
            self.import_patterns["pydantic"].pop("Field")
            self.import_patterns["xsdata_pydantic_basemodel.pydantic_compat"] = {
                "Field": {" = Field("}
            }

    @classmethod
    def build_import_patterns(cls) -> dict[str, dict]:
        patterns = Filters.build_import_patterns()
        patterns.update(
            {"pydantic": {"Field": [" = Field("], "BaseModel": ["BaseModel"]}}
        )
        return {key: patterns[key] for key in sorted(patterns)}

    @classmethod
    def build_class_annotation(cls, fmt: OutputFormat) -> str:
        # remove the @dataclass decorator
        return ""

    def field_definition(
        self, attr: Attr, ns_map: dict, parent_namespace: str | None, parents: list[str]
    ) -> str:
        defn = super().field_definition(attr, ns_map, parent_namespace, parents)
        return defn.replace("field(", "Field(")

    def format_arguments(self, kwargs: dict, indent: int = 0) -> str:
        # called by field_definition
        self.move_restrictions_to_pydantic_field(kwargs)
        return super().format_arguments(kwargs, indent)

    def class_bases(self, obj: Class, class_name: str) -> list[str]:
        # add BaseModel to the class bases
        # FIXME ... need to dedupe superclasses
        bases = super().class_bases(obj, class_name)
        return unique_sequence([*bases, "BaseModel"])

    def move_restrictions_to_pydantic_field(
        self, kwargs: dict, pop: bool = False
    ) -> None:
        """Move metadata from the metadata dict to the pydantic Field kwargs."""
        # XXX: can we pop them?  or does xsdata need them in the metadata dict as well?
        if "metadata" not in kwargs:  # pragma: no cover
            return

        # The choice to use v1 syntax for cross-compatible mode has to do with
        # https://docs.pydantic.dev/usage/schema/#unenforced-field-constraints
        # There were more fields in v1 than in v2, so "min_length" is degenerate in v2
        # NOTE: ... this might be fixed by using pydantic_compat?
        if self.pydantic_support == "v2":
            use_v2 = True
        elif self.pydantic_support == "auto":
            use_v2 = PYDANTIC2
        else:  # v1 or both
            use_v2 = False

        restriction_map = V2_RESTRICTION_MAP if use_v2 else V1_RESTRICTION_MAP

        metadata: dict = kwargs["metadata"]
        getitem = metadata.pop if pop else metadata.get
        for from_, to_ in restriction_map.items():
            if from_ in metadata:
                kwargs[to_] = getitem(from_)

        if use_v2 and "metadata" in kwargs:
            kwargs["json_schema_extra"] = kwargs.pop("metadata")

    # note, this method is the same as the base class implementation before it
    # was changed in xsdata v24.  that change breaks private package names... so
    # we're using the old implementation here.
    # https://github.com/tefra/xsdata/issues/948
    def safe_name(
        self, name: str, prefix: str, name_case: Callable, **kwargs: Any
    ) -> str:
        """Sanitize names for safe generation."""
        if not name:
            return self.safe_name(prefix, prefix, name_case, **kwargs)

        if re.match(r"^-\d*\.?\d+$", name):
            return self.safe_name(f"{prefix}_minus_{name}", prefix, name_case, **kwargs)

        slug = text.alnum(name)
        if not slug or not slug[0].isalpha():
            return self.safe_name(f"{prefix}_{name}", prefix, name_case, **kwargs)

        result = name_case(name, **kwargs)
        if text.is_reserved(result):
            return self.safe_name(f"{name}_{prefix}", prefix, name_case, **kwargs)

        return result


V1_RESTRICTION_MAP = {
    "min_occurs": "min_items",  # semantics are different
    "max_occurs": "max_items",  # semantics are different
    "min_exclusive": "gt",
    "min_inclusive": "ge",
    "max_exclusive": "lt",
    "max_inclusive": "le",
    "min_length": "min_length",  # only applies to strings
    "max_length": "max_length",  # only applies to strings
    "pattern": "regex",
    "fraction_digits": "decimal_places",
    "total_digits": "max_digits",
    # --- other restrictions that don't have a direct mapping ---
    # "length": "...",
    # "white_space": "...",
    # "explicit_timezone": "...",
    # "nillable": "...",
    # "sequence": "...",
    # "tokens": "...",
    # "format": "...",
    # "choice": "...",
    # "group": "...",
    # "path": "...",
}
V2_RESTRICTION_MAP = {
    **V1_RESTRICTION_MAP,
    "min_occurs": "min_length",  # semantics are different
    "max_occurs": "max_length",  # semantics are different
    "pattern": "pattern",
}



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, cast

from xsdata.formats.dataclass.filters import Filters
from xsdata.formats.dataclass.generator import DataclassGenerator

import ome_autogen.overrides as ovr
from ome_autogen import _util
from xsdata_pydantic_basemodel.generator import PydanticBaseFilters

if TYPE_CHECKING:
    from jinja2 import Environment, FileSystemLoader
    from xsdata.codegen.models import Attr, Class
    from xsdata.codegen.resolver import DependenciesResolver
    from xsdata.models.config import GeneratorConfig


class OmeGenerator(DataclassGenerator):
    @classmethod
    def init_filters(cls, config: GeneratorConfig) -> Filters:
        return OmeFilters(config)

    def render_module(
        self, resolver: DependenciesResolver, classes: list[Class]
    ) -> str:
        mod = super().render_module(resolver, classes)

        # xsdata renames classes like "FillRule" (which appears as a SimpleType
        # inside of the Shape ComplexType) as "Shape_FillRule".
        # We want to make them available as "FillRule" in the corresponding
        # module, (i.e. the "Shape" module in this case).
        # That is, we want "Shape = Shape_FillRule" included in the module.
        # this is for backwards compatibility.
        aliases = []
        for c in classes:
            for i in resolver.imports:
                if f"{c.name}_" in i.qname:
                    # import_name is something like 'Shape_FillRule'
                    import_name = i.qname.rsplit("}", 1)[-1]
                    # desired alias is just 'FillRule'
                    alias = import_name.split(f"{c.name}_")[-1]
                    aliases.append(f"{alias} = {import_name}")

            # we also want inner (nested) classes to be available at the top level
            # e.g. map.Map.M -> map.M
            for inner in c.inner:
                aliases.append(f"{inner.name} = {c.name}.{inner.name}")

        if aliases:
            mod += "\n\n" + "\n".join(aliases) + "\n"

        return mod


class OmeFilters(PydanticBaseFilters):
    def register(self, env: Environment) -> None:
        # add our own templates dir to the search path
        tpl_dir = Path(__file__).parent.joinpath("templates")
        cast("FileSystemLoader", env.loader).searchpath.insert(0, str(tpl_dir))
        super().register(env)
        env.filters.update({"methods": self.methods})

    def __init__(self, config: GeneratorConfig):
        super().__init__(config)

        # TODO: it would be nice to know how to get the schema we're processing from
        # the config.  For now, we just assume it's the OME schema and that's the
        # hardcoded default in _util.get_appinfo
        self.appinfo = _util.get_appinfo()

    @contextmanager
    def _modern_typing(self) -> Iterator[None]:
        """Context manager to use modern typing syntax."""
        prev_u, self.union_type = self.union_type, True
        prev_s, self.subscriptable_types = self.subscriptable_types, True
        try:
            yield
        finally:
            self.union_type = prev_u
            self.subscriptable_types = prev_s

    def class_params(self, obj: Class) -> Iterator[tuple[str, str, str]]:  # type: ignore[override]
        # This method override goes along with the docstring jinja template override
        # to fixup the numpy docstring format.
        # https://github.com/tefra/xsdata/issues/818
        # The type ignore is because xsdata returns an iterator of 2-tuples
        # but we want to return a 3-tuple.
        for attr in obj.attrs:
            # IMPORTANT:
            # self.field_type must be called before getting attr.name, since that
            # is currently where plural names are corrected.  (This is a hack.)
            with self._modern_typing():
                type_ = self.field_type(attr, [obj.name])

            name = (
                self.constant_name(attr.name, obj.name)
                if obj.is_enumeration
                else self.field_name(attr.name, obj.name)
            )
            if not (help_ := attr.help):
                help_ = f"(The {obj.name} {attr.name})."
            yield name, type_, self.clean_docstring(help_)

    def class_bases(self, obj: Class, class_name: str) -> list[str]:
        # we don't need PydanticBaseFilters to add the Base class
        # because we add it in the config.extensions
        # This could go once PydanticBaseFilters is better about deduping
        return Filters.class_bases(self, obj, class_name)

    def _attr_is_optional(self, attr: Attr) -> bool:
        if attr.name in ovr.NEVER_OPTIONAL:
            return False
        return attr.is_nillable or (
            attr.default is None and (attr.is_optional or not self.format.kw_only)
        )

    def _format_type(self, attr: Attr, result: str) -> str:
        if self._attr_is_optional(attr):
            return f"None | {result}" if self.union_type else f"Optional[{result}]"
        return result

    def field_type(self, attr: Attr, parents: list[str]) -> str:
        if attr.is_list and not getattr(attr, "_plural_set", False):
            # HACK
            # It would be nicer to put this in the self.field_name method...but that
            # method only receives the attr name, not the attr object, and so we
            # don't know at that point whether it belongs to a list or not.
            # This hack works only because this method is called BEFORE self.field_name
            # in the class.jinja2 template, so we directly modify the attr object here.
            attr.name = self.appinfo.plurals.get(attr.name, f"{attr.name}s")
            attr._plural_set = True  # type: ignore

        if attr.name in ovr.OVERRIDE_ELEM_TO_CLASS:
            return self._format_type(attr, ovr.OVERRIDE_ELEM_TO_CLASS[attr.name])

        type_name = super().field_type(attr, parents)
        # we want to use datetime.datetime instead of XmlDateTime
        return type_name.replace("XmlDateTime", "datetime")

    @classmethod
    def build_import_patterns(cls) -> dict[str, dict]:
        import_patterns = super().build_import_patterns()
        for mod, types in ovr.IMPORT_PATTERNS.items():
            mod_entry: dict[str, list[str]] = import_patterns.setdefault(mod, {})
            for name, patterns in types.items():
                name_entry: list[str] = mod_entry.setdefault(name, [])
                name_entry.extend(patterns)
        return {key: import_patterns[key] for key in sorted(import_patterns)}

    def field_default_value(self, attr: Attr, ns_map: dict | None = None) -> str:
        if override := ovr.get(attr.name):
            if override.default_factory:
                return override.default_factory
            if override.default:
                return override.default

        return super().field_default_value(attr, ns_map)

    def format_arguments(self, kwargs: dict, indent: int = 0) -> str:
        # keep default_factory at the front
        attr_name = kwargs["metadata"].get("name")
        if (override := ovr.get(attr_name)) and override.default_factory:
            kwargs.pop("default", None)
            kwargs["default_factory"] = override.default_factory

        # uncomment this to use new_uuid as the default_factory for all UUIDs
        # but then we have an equality checking problem in the tests
        # if kwargs.get("metadata", {}).get("pattern", "").startswith("(urn:uuid:"):
        #     kwargs.pop("default", None)
        #     kwargs = {"default_factory": "new_uuid", **kwargs}
        return super().format_arguments(kwargs, indent)

    def constant_name(self, name: str, class_name: str) -> str:
        if class_name in self.appinfo.enums:
            # use the enum names found in appinfo/xsdfu/enum
            return self.appinfo.enums[class_name][name].enum
        return super().constant_name(name, class_name)

    def methods(self, obj: Class) -> list[str]:
        if (override := ovr.get(obj.name)) and (lines := override.add_lines):
            return ["\n\n" + "\n".join(lines)]
        return []



from dataclasses import dataclass
from typing import Literal

from xsdata.models import config


@dataclass
class GeneratorOutput(config.GeneratorOutput):
    # v1 will only support pydantic<2
    # v2 will only support pydantic>=2
    # auto will only support whatever pydantic is installed at codegen time
    # both will support both pydantic versions
    pydantic_support: Literal["v1", "v2", "auto", "both"] = "auto"

    def __post_init__(self) -> None:
        if self.pydantic_support not in ("v1", "v2", "auto", "both"):
            raise ValueError(
                "pydantic_support must be one of 'v1', 'v2', 'auto', 'both', not "
                f"{self.pydantic_support!r}"
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from ome_types._conversion import OME_2016_06_URI

if TYPE_CHECKING:
    from ome_types._conversion import AnyElementTree


NSMAP = {"": OME_2016_06_URI, "ome": OME_2016_06_URI}


# See note before using...
def fix_micro_manager_instrument(tree: AnyElementTree) -> AnyElementTree:
    """Fix MicroManager Instrument and Detector IDs and References.

    Some versions of OME-XML produced by MicroManager have invalid IDs (and references)
    for Instruments and Detectors. This function fixes those IDs and references.

    NOTE: as of v0.4.0, bad IDs and references are caught during ID validation anyway,
    so this is mostly an example of a fix function, and could be used to prevent
    the warning from being raised.
    """
    for i_idx, instrument in enumerate(tree.findall("Instrument", NSMAP)):
        old_id = instrument.get("ID")  # type: ignore [attr-defined]
        if old_id.startswith("Microscope"):
            new_id = f"Instrument:{i_idx}"
            instrument.set("ID", new_id)  # type: ignore [attr-defined]
            for ref in tree.findall(f".//InstrumentRef[@ID='{old_id}']", NSMAP):
                ref.set("ID", new_id)

        detectors = instrument.findall(".//Detector", NSMAP)  # type: ignore [attr-defined]
        for d_idx, detector in enumerate(detectors):
            old_id = detector.get("ID")
            if not old_id.startswith("Detector:"):
                new_id = f"Detector:{old_id if old_id.isdigit() else d_idx}"
                detector.set("ID", new_id)
                for ref in tree.findall(f".//DetectorSettings[@ID='{old_id}']", NSMAP):
                    ref.set("ID", new_id)

    return tree


ALL_FIXES = [fix_micro_manager_instrument]
__all__ = ["ALL_FIXES", "fix_micro_manager_instrument"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, MutableSequence, cast

import pydantic.version
from pydantic import BaseModel

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo

pydantic_version: tuple[int, ...] = tuple(
    int(x) for x in pydantic.version.VERSION.split(".")[:2]
)


if pydantic_version >= (2,):
    try:
        from pydantic_extra_types.color import Color as Color
    except ImportError:
        from pydantic.color import Color as Color

    def field_type(field: FieldInfo) -> Any:
        return field.annotation

    def field_regex(obj: type[BaseModel], field_name: str) -> str | None:
        # typing is incorrect at the moment, but may indicate breakage in pydantic 3
        field_info = obj.model_fields[field_name]  # type: ignore [index]
        meta = field_info.json_schema_extra or {}
        # if a "metadata" key exists... use it.
        # After pydantic-compat 0.2, this is where it will be.
        if "metadata" in meta:  # type: ignore
            meta = meta["metadata"]  # type: ignore
        if meta:
            return meta.get("pattern")  # type: ignore
        return None

    kw: dict = {"validated_data": {}} if pydantic_version >= (2, 10) else {}

    def get_default(f: FieldInfo) -> Any:
        return f.get_default(call_default_factory=True, **kw)
else:
    from pydantic.color import Color as Color  # type: ignore [no-redef]

    def field_type(field: Any) -> Any:  # type: ignore
        return field.type_

    def field_regex(obj: type[BaseModel], field_name: str) -> str | None:
        field = obj.__fields__[field_name]  # type: ignore
        return cast(str, field.field_info.regex)

    def get_default(f: Any) -> Any:  # type: ignore
        return f.get_default()


def update_set_fields(self: BaseModel) -> None:
    """Update set fields with populated mutable sequences.

    Because pydantic isn't aware of mutations to sequences, it can't tell when
    a field has been "set" by mutating a sequence.  This method updates the
    self.__fields_set__ attribute to reflect that.  We assume that if an attribute
    is not None, and is not equal to the default value, then it has been set.
    """
    for field_name, field in self.model_fields.items():
        current = getattr(self, field_name)
        if not current:
            continue
        if current != get_default(field):
            self.model_fields_set.add(field_name)
        if isinstance(current, BaseModel):
            update_set_fields(current)
        if isinstance(current, MutableSequence):
            for item in current:
                if isinstance(item, BaseModel):
                    update_set_fields(item)



import dataclasses as dc
from contextlib import suppress
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
)

try:
    from lxml import etree as ET
except ImportError:
    import xml.etree.ElementTree as ET  # type: ignore

from pydantic import BaseModel, validators
from pydantic_compat import PYDANTIC2, PydanticCompatMixin
from xsdata.formats.dataclass.compat import Dataclasses, class_types
from xsdata.formats.dataclass.models.elements import XmlType
from xsdata.models.datatype import XmlDate, XmlDateTime, XmlDuration, XmlPeriod, XmlTime

from xsdata_pydantic_basemodel.pydantic_compat import Field, dataclass_fields

T = TypeVar("T", bound=object)

if TYPE_CHECKING:
    from pydantic import ConfigDict


class AnyElement(PydanticCompatMixin, BaseModel):
    """Generic model to bind xml document data to wildcard fields.

    :param qname: The element's qualified name
    :param text: The element's text content
    :param tail: The element's tail content
    :param children: The element's list of child elements.
    :param attributes: The element's key-value attribute mappings.
    """

    qname: Optional[str] = Field(default=None)
    text: Optional[str] = Field(default=None)
    tail: Optional[str] = Field(default=None)
    children: List["AnyElement"] = Field(
        default_factory=list,
        metadata={"type": XmlType.WILDCARD},  # type: ignore [call-arg,call-overload]
    )
    attributes: Dict[str, str] = Field(
        default_factory=dict,
        metadata={"type": XmlType.ATTRIBUTES},  # type: ignore [call-arg,call-overload]
    )

    model_config: ClassVar["ConfigDict"] = {"arbitrary_types_allowed": True}

    def to_etree_element(self) -> "ET._Element":
        elem = ET.Element(self.qname or "", self.attributes)
        elem.text = self.text
        elem.tail = self.tail
        for child in self.children:
            elem.append(child.to_etree_element())
        return elem


class DerivedElement(PydanticCompatMixin, BaseModel, Generic[T]):
    """Generic model wrapper for type substituted elements.

    Example: eg. <b xsi:type="a">...</b>

    :param qname: The element's qualified name
    :param value: The wrapped value
    :param type: The real xsi:type
    """

    qname: str
    value: T
    type: Optional[str] = None

    model_config: ClassVar["ConfigDict"] = {"arbitrary_types_allowed": True}


class PydanticBaseModel(Dataclasses):
    @property
    def any_element(self) -> Type:
        return AnyElement

    @property
    def derived_element(self) -> Type:
        return DerivedElement

    def is_model(self, obj: Any) -> bool:
        with suppress(Exception):
            clazz = obj if isinstance(obj, type) else type(obj)
            if issubclass(clazz, BaseModel) and clazz != BaseModel:
                rebuild = getattr(clazz, "model_rebuild", None)
                if callable(rebuild):
                    rebuild()
                return True
        return False

    #
    # https://github.com/tefra/xsdata/pull/949
    def get_fields(self, obj: Any) -> Iterator[dc.Field]:  # type: ignore[override]
        yield from dataclass_fields(obj)


class_types.register("pydantic-basemodel", PydanticBaseModel())


def make_validators(tp: Type, factory: Callable) -> List[Callable]:
    def validator(value: Any) -> Any:
        if isinstance(value, tp):
            return value

        if isinstance(value, str):
            return factory(value)

        raise ValueError

    return [validator]


_validators = {
    XmlDate: make_validators(XmlDate, XmlDate.from_string),
    XmlDateTime: make_validators(XmlDateTime, XmlDateTime.from_string),
    XmlTime: make_validators(XmlTime, XmlTime.from_string),
    XmlDuration: make_validators(XmlDuration, XmlDuration),
    XmlPeriod: make_validators(XmlPeriod, XmlPeriod),
    ET.QName: make_validators(ET.QName, ET.QName),
}

if not PYDANTIC2:
    validators._VALIDATORS.extend(list(_validators.items()))
else:
    from pydantic import BaseModel
    from pydantic_core import core_schema as cs

    def _make_get_core_schema(validator: Callable) -> Callable:
        def get_core_schema(*args: Any) -> cs.PlainValidatorFunctionSchema:
            return cs.general_plain_validator_function(validator)

        return get_core_schema

    for type_, val in _validators.items():
        get_schema = _make_get_core_schema(val[0])
        with suppress(TypeError):
            type_.__get_pydantic_core_schema__ = get_schema  # type: ignore



import itertools
from typing import Any, Generic, Iterator, List, TypeVar, Union, cast, no_type_check

from pydantic import BaseModel

# for circular import reasons...
from ome_types._autogenerated.ome_2016_06.boolean_annotation import BooleanAnnotation
from ome_types._autogenerated.ome_2016_06.comment_annotation import CommentAnnotation
from ome_types._autogenerated.ome_2016_06.double_annotation import DoubleAnnotation

# for circular import reasons...
from ome_types._autogenerated.ome_2016_06.ellipse import Ellipse
from ome_types._autogenerated.ome_2016_06.file_annotation import FileAnnotation
from ome_types._autogenerated.ome_2016_06.label import Label
from ome_types._autogenerated.ome_2016_06.line import Line
from ome_types._autogenerated.ome_2016_06.list_annotation import ListAnnotation
from ome_types._autogenerated.ome_2016_06.long_annotation import LongAnnotation
from ome_types._autogenerated.ome_2016_06.map_annotation import MapAnnotation
from ome_types._autogenerated.ome_2016_06.mask import Mask
from ome_types._autogenerated.ome_2016_06.point import Point
from ome_types._autogenerated.ome_2016_06.polygon import Polygon
from ome_types._autogenerated.ome_2016_06.polyline import Polyline
from ome_types._autogenerated.ome_2016_06.rectangle import Rectangle
from ome_types._autogenerated.ome_2016_06.tag_annotation import TagAnnotation
from ome_types._autogenerated.ome_2016_06.term_annotation import TermAnnotation
from ome_types._autogenerated.ome_2016_06.timestamp_annotation import (
    TimestampAnnotation,
)
from ome_types._autogenerated.ome_2016_06.xml_annotation import XMLAnnotation

T = TypeVar("T")


class CollectionMixin(BaseModel, Generic[T]):
    """Mixin to be used for classes that behave like collections.

    Notably: ROI.Union and StructuredAnnotations.
    All the fields in these types list[SomeType], and they collectively behave like
    a list with the union of all field types.
    """

    @no_type_check
    def __iter__(self) -> Iterator[T]:
        return itertools.chain(*(getattr(self, f) for f in self.model_fields))

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def append(self, item: T) -> None:
        """Append an item to the appropriate field list."""
        cast(list, getattr(self, self._field_name(item))).append(item)

    def extend(self, items: List[T]) -> None:
        """Extend the appropriate field list with the given items."""
        for item in items:
            self.append(item)

    def remove(self, item: T) -> None:
        """Remove an item from the appropriate field list."""
        cast(list, getattr(self, self._field_name(item))).remove(item)

    # This one is a bit hacky... perhaps deprecate and remove
    def __getitem__(self, i: int) -> T:
        # return the ith item in the __iter__ sequence
        return next(itertools.islice(self, i, None))

    # perhaps deprecate and remove
    def __eq__(self, _value: object) -> bool:
        if isinstance(_value, list):
            return list(self) == _value
        return super().__eq__(_value)

    @classmethod
    def _field_name(cls, item: T) -> str:
        """Return the name of the field that should contain the given item.

        Must be implemented by subclasses.
        """
        raise NotImplementedError()  # pragma: no cover


# ------------------------ StructuredAnnotations ------------------------

AnnotationType = Union[
    XMLAnnotation,
    FileAnnotation,
    ListAnnotation,
    LongAnnotation,
    DoubleAnnotation,
    CommentAnnotation,
    BooleanAnnotation,
    TimestampAnnotation,
    TagAnnotation,
    TermAnnotation,
    MapAnnotation,
]
# get_args wasn't available until Python 3.8
AnnotationInstances = AnnotationType.__args__  # type: ignore


class StructuredAnnotationsMixin(CollectionMixin[AnnotationType]):
    @classmethod
    def _field_name(cls, item: Any) -> str:
        if not isinstance(item, AnnotationInstances):
            raise TypeError(  # pragma: no cover
                f"Expected an instance of {AnnotationInstances}, got {item!r}"
            )
        # where 10 is the length of "Annotation"
        return item.__class__.__name__[:-10].lower() + "_annotations"


ShapeType = Union[Rectangle, Mask, Point, Ellipse, Line, Polyline, Polygon, Label]
ShapeInstances = ShapeType.__args__  # type: ignore


class ShapeUnionMixin(CollectionMixin[ShapeType]):
    @classmethod
    def _field_name(cls, item: Any) -> str:
        if not isinstance(item, ShapeInstances):
            raise TypeError(  # pragma: no cover
                f"Expected an instance of {ShapeInstances}, got {item!r}"
            )
        return item.__class__.__name__.lower() + "s"



from __future__ import annotations

import dataclasses as dc
import warnings
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from pydantic_compat import PYDANTIC2, Field

__all__ = ["Field", "PYDANTIC2"]

if TYPE_CHECKING:
    from pydantic import BaseModel

    M = TypeVar("M", bound=BaseModel)
    C = TypeVar("C", bound=Callable[..., Any])


if PYDANTIC2:
    from pydantic.fields import FieldInfo
    from pydantic_core import PydanticUndefined as Undefined

    def _get_metadata(pydantic_field: FieldInfo) -> dict[str, Any]:
        meta = (
            pydantic_field.json_schema_extra
            if isinstance(pydantic_field.json_schema_extra, dict)
            else {}
        )
        # if a "metadata" key exists... use it.
        # After pydantic-compat 0.2, this is where it will be.
        if "metadata" in meta:
            meta = meta["metadata"]  # type: ignore
        return meta

else:
    from pydantic.fields import Undefined as Undefined  # type: ignore

    def _get_metadata(pydantic_field) -> dict:  # type: ignore
        extra = pydantic_field.field_info.extra
        if "json_schema_extra" in extra:
            return extra["json_schema_extra"]
        return extra.get("metadata", {})


def _get_defaults(pydantic_field: FieldInfo) -> tuple[Any, Any]:
    if pydantic_field.default_factory is not None:
        default_factory: Any = pydantic_field.default_factory
        default = dc.MISSING
    else:
        default_factory = dc.MISSING
        default = (
            dc.MISSING
            if pydantic_field.default in (Undefined, Ellipsis)
            else pydantic_field.default
        )
    return default_factory, default


def _pydantic_field_to_dataclass_field(
    name: str, pydantic_field: FieldInfo
) -> dc.Field:
    default_factory, default = _get_defaults(pydantic_field)

    metadata = _get_metadata(pydantic_field).copy()

    # HACK
    # see https://github.com/tlambert03/ome-types/pull/235 for description of problem
    # This is a hack to get around the fact that xsdata requires Element choices
    # to be added to the Field metadata as `choices: List[dict]` ... but pydantic
    # requires that everything in a Field be hashable (if you want to cast the model
    # to a JSON schema), and `dict` is not hashable. So here, when we're converting
    # a pydantic Field to a dataclass Field for xsdata to consume, we cast all items
    # in the `choices` list to `dict` (which is hashable).
    # Then, in our source code, we declare choices as tuple[tuple[str, str], ...]
    # which IS hashable.
    if "choices" in metadata:
        choices = []
        for choice in metadata["choices"]:
            choice = dict(choice)
            # we also, unfortunately, need to convert the "type" field from a
            # class name to a class object
            if "type" in choice and isinstance(choice["type"], str):
                try:
                    from ome_types import model

                    choice["type"] = getattr(model, choice["type"])
                except AttributeError:
                    warnings.warn(
                        f"Could not find {choice['type']} in ome_types.model",
                        stacklevel=2,
                    )

            choices.append(choice)
        metadata["choices"] = choices

    dataclass_field = dc.field(  # type: ignore
        default=default, default_factory=default_factory, metadata=metadata
    )
    dataclass_field.name = name
    return dataclass_field


@lru_cache(maxsize=None)
def dataclass_fields(obj: type[M]) -> tuple[dc.Field, ...]:
    """Return a tuple of dataclass fields for the given pydantic model class."""
    return tuple(
        _pydantic_field_to_dataclass_field(name, f)
        for name, f in obj.model_fields.items()  # type: ignore
    )



"""
Vendored from
https://github.com/pydantic/pydantic-extra-types/blob/0c42c50c9d8c6ee77b7e9fd3ca437856b2305eb0/pydantic_extra_types/color.py
The MIT License (MIT)

Copyright (c) 2023 Samuel Colvin and other contributors


Color definitions are used as per the CSS3
[CSS Color Module Level 3](http://www.w3.org/TR/css3-color/#svg-color) specification.

A few colors have multiple names referring to the sames colors, eg. `grey` and `gray` or `aqua` and `cyan`.

In these cases the _last_ color when sorted alphabetically takes preferences,
eg. `Color((0, 255, 255)).as_named() == 'cyan'` because "cyan" comes after "aqua".
"""

from __future__ import annotations

import math
import re
import sys
import typing
from colorsys import hls_to_rgb, rgb_to_hls
from typing import TYPE_CHECKING, Any, Callable, Tuple, Union, cast, Literal

from pydantic_core import CoreSchema, PydanticCustomError, core_schema

ColorTuple = Union[Tuple[int, int, int], Tuple[int, int, int, float]]
ColorType = Union[ColorTuple, str, "Color"]
HslColorTuple = Union[Tuple[float, float, float], Tuple[float, float, float, float]]

if TYPE_CHECKING:
    from pydantic import GetJsonSchemaHandler
    from pydantic.json_schema import JsonSchemaValue


class RGBA:
    """
    Internal use only as a representation of a color.
    """

    __slots__ = "r", "g", "b", "alpha", "_tuple"

    def __init__(self, r: float, g: float, b: float, alpha: float | None):
        self.r = r
        self.g = g
        self.b = b
        self.alpha = alpha

        self._tuple: tuple[float, float, float, float | None] = (r, g, b, alpha)

    def __getitem__(self, item: Any) -> Any:
        return self._tuple[item]


# these are not compiled here to avoid import slowdown, they'll be compiled the first time they're used, then cached
_r_255 = r"(\d{1,3}(?:\.\d+)?)"
_r_comma = r"\s*,\s*"
_r_alpha = r"(\d(?:\.\d+)?|\.\d+|\d{1,2}%)"
_r_h = r"(-?\d+(?:\.\d+)?|-?\.\d+)(deg|rad|turn)?"
_r_sl = r"(\d{1,3}(?:\.\d+)?)%"
r_hex_short = r"\s*(?:#|0x)?([0-9a-f])([0-9a-f])([0-9a-f])([0-9a-f])?\s*"
r_hex_long = r"\s*(?:#|0x)?([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})?\s*"
# CSS3 RGB examples: rgb(0, 0, 0), rgba(0, 0, 0, 0.5), rgba(0, 0, 0, 50%)
r_rgb = rf"\s*rgba?\(\s*{_r_255}{_r_comma}{_r_255}{_r_comma}{_r_255}(?:{_r_comma}{_r_alpha})?\s*\)\s*"
# CSS3 HSL examples: hsl(270, 60%, 50%), hsla(270, 60%, 50%, 0.5), hsla(270, 60%, 50%, 50%)
r_hsl = rf"\s*hsla?\(\s*{_r_h}{_r_comma}{_r_sl}{_r_comma}{_r_sl}(?:{_r_comma}{_r_alpha})?\s*\)\s*"
# CSS4 RGB examples: rgb(0 0 0), rgb(0 0 0 / 0.5), rgb(0 0 0 / 50%), rgba(0 0 0 / 50%)
r_rgb_v4_style = (
    rf"\s*rgba?\(\s*{_r_255}\s+{_r_255}\s+{_r_255}(?:\s*/\s*{_r_alpha})?\s*\)\s*"
)
# CSS4 HSL examples: hsl(270 60% 50%), hsl(270 60% 50% / 0.5), hsl(270 60% 50% / 50%), hsla(270 60% 50% / 50%)
r_hsl_v4_style = (
    rf"\s*hsla?\(\s*{_r_h}\s+{_r_sl}\s+{_r_sl}(?:\s*/\s*{_r_alpha})?\s*\)\s*"
)

# colors where the two hex characters are the same, if all colors match this the short version of hex colors can be used
repeat_colors = {int(c * 2, 16) for c in "0123456789abcdef"}
rads = 2 * math.pi


class Representation:
    # Mixin to provide `__str__`, `__repr__`, and `__pretty__` and `__rich_repr__` methods.
    # `__pretty__` is used by [devtools](https://python-devtools.helpmanual.io/).
    # `__rich_repr__` is used by [rich](https://rich.readthedocs.io/en/stable/pretty.html).
    # (this is not a docstring to avoid adding a docstring to classes which inherit from Representation)

    # we don't want to use a type annotation here as it can break get_type_hints
    __slots__ = tuple()  # type: typing.Collection[str]

    def __repr_args__(self) -> Any:
        """Returns the attributes to show in __str__, __repr__, and __pretty__ this is generally overridden.

        Can either return:
        * name - value pairs, e.g.: `[('foo_name', 'foo'), ('bar_name', ['b', 'a', 'r'])]`
        * or, just values, e.g.: `[(None, 'foo'), (None, ['b', 'a', 'r'])]`
        """
        attrs_names = self.__slots__
        if not attrs_names and hasattr(self, "__dict__"):
            attrs_names = self.__dict__.keys()
        attrs = ((s, getattr(self, s)) for s in attrs_names)
        return [(a, v) for a, v in attrs if v is not None]

    def __repr_name__(self) -> str:
        """Name of the instance's class, used in __repr__."""
        return self.__class__.__name__

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}" for a, v in self.__repr_args__()
        )

    def __pretty__(
        self, fmt: typing.Callable[[Any], Any], **kwargs: Any
    ) -> typing.Generator[Any, None, None]:
        """Used by devtools (https://python-devtools.helpmanual.io/) to pretty print objects."""
        yield self.__repr_name__() + "("
        yield 1
        for name, value in self.__repr_args__():
            if name is not None:
                yield name + "="
            yield fmt(value)
            yield ","
            yield 0
        yield -1
        yield ")"

    def __rich_repr__(self) -> Any:
        """Used by Rich (https://rich.readthedocs.io/en/stable/pretty.html) to pretty print objects."""
        for name, field_repr in self.__repr_args__():
            if name is None:
                yield field_repr
            else:
                yield name, field_repr

    def __str__(self) -> str:
        return self.__repr_str__(" ")

    def __repr__(self) -> str:
        return f'{self.__repr_name__()}({self.__repr_str__(", ")})'


class Color(Representation):
    """
    Represents a color.
    """

    __slots__ = "_original", "_rgba"

    def __init__(self, value: ColorType) -> None:
        self._rgba: RGBA
        self._original: ColorType
        if isinstance(value, (tuple, list)):
            self._rgba = parse_tuple(value)
        elif isinstance(value, str):
            self._rgba = parse_str(value)
        elif isinstance(value, Color):
            self._rgba = value._rgba
            value = value._original
        else:
            raise PydanticCustomError(
                "color_error",
                "value is not a valid color: value must be a tuple, list or string",
            )

        # if we've got here value must be a valid color
        self._original = value

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        field_schema: dict[str, Any] = {}
        field_schema.update(type="string", format="color")
        return field_schema

    def original(self) -> ColorType:
        """
        Original value passed to `Color`.
        """
        return self._original

    def as_named(self, *, fallback: bool = False) -> str:
        """
        Returns the name of the color if it can be found in `COLORS_BY_VALUE` dictionary,
        otherwise returns the hexadecimal representation of the color or raises `ValueError`.

        Args:
        ----
            fallback: If True, falls back to returning the hexadecimal representation of
                the color instead of raising a ValueError when no named color is found.

        Returns:
        -------
            The name of the color, or the hexadecimal representation of the color.

        Raises:
        ------
            ValueError: When no named color is found and fallback is `False`.
        """
        if self._rgba.alpha is None:
            rgb = cast(Tuple[int, int, int], self.as_rgb_tuple())
            try:
                return COLORS_BY_VALUE[rgb]
            except KeyError as e:
                if fallback:
                    return self.as_hex()
                else:
                    raise ValueError(
                        "no named color found, use fallback=True, as_hex() or as_rgb()"
                    ) from e
        else:
            return self.as_hex()

    def as_hex(self, format: Literal["short", "long"] = "short") -> str:
        """Returns the hexadecimal representation of the color.

        Hex string representing the color can be 3, 4, 6, or 8 characters depending on whether the string
        a "short" representation of the color is possible and whether there's an alpha channel.

        Returns
        -------
            The hexadecimal representation of the color.
        """
        values = [float_to_255(c) for c in self._rgba[:3]]
        if self._rgba.alpha is not None:
            values.append(float_to_255(self._rgba.alpha))

        as_hex = "".join(f"{v:02x}" for v in values)
        if format == "short" and all(c in repeat_colors for c in values):
            as_hex = "".join(as_hex[c] for c in range(0, len(as_hex), 2))
        return "#" + as_hex

    def as_rgb(self) -> str:
        """
        Color as an `rgb(<r>, <g>, <b>)` or `rgba(<r>, <g>, <b>, <a>)` string.
        """
        if self._rgba.alpha is None:
            return f"rgb({float_to_255(self._rgba.r)}, {float_to_255(self._rgba.g)}, {float_to_255(self._rgba.b)})"
        else:
            return (
                f"rgba({float_to_255(self._rgba.r)}, {float_to_255(self._rgba.g)}, {float_to_255(self._rgba.b)}, "
                f"{round(self._alpha_float(), 2)})"
            )

    def as_rgb_tuple(self, *, alpha: bool | None = None) -> ColorTuple:
        """
        Returns the color as an RGB or RGBA tuple.

        Args:
        ----
            alpha: Whether to include the alpha channel. There are three options for this input:

                - `None` (default): Include alpha only if it's set. (e.g. not `None`)
                - `True`: Always include alpha.
                - `False`: Always omit alpha.

        Returns:
        -------
            A tuple that contains the values of the red, green, and blue channels in the range 0 to 255.
                If alpha is included, it is in the range 0 to 1.
        """
        r, g, b = (float_to_255(c) for c in self._rgba[:3])
        if alpha is None:
            if self._rgba.alpha is None:
                return r, g, b
            else:
                return r, g, b, self._alpha_float()
        elif alpha:
            return r, g, b, self._alpha_float()
        else:
            # alpha is False
            return r, g, b

    def as_hsl(self) -> str:
        """
        Color as an `hsl(<h>, <s>, <l>)` or `hsl(<h>, <s>, <l>, <a>)` string.
        """
        if self._rgba.alpha is None:
            h, s, li = self.as_hsl_tuple(alpha=False)  # type: ignore
            return f"hsl({h * 360:0.0f}, {s:0.0%}, {li:0.0%})"
        else:
            h, s, li, a = self.as_hsl_tuple(alpha=True)  # type: ignore
            return f"hsl({h * 360:0.0f}, {s:0.0%}, {li:0.0%}, {round(a, 2)})"

    def as_hsl_tuple(self, *, alpha: bool | None = None) -> HslColorTuple:
        """
        Returns the color as an HSL or HSLA tuple.

        Args:
        ----
            alpha: Whether to include the alpha channel.

                - `None` (default): Include the alpha channel only if it's set (e.g. not `None`).
                - `True`: Always include alpha.
                - `False`: Always omit alpha.

        Returns:
        -------
            The color as a tuple of hue, saturation, lightness, and alpha (if included).
                All elements are in the range 0 to 1.

        Note:
        ----
            This is HSL as used in HTML and most other places, not HLS as used in Python's `colorsys`.
        """
        h, l, s = rgb_to_hls(self._rgba.r, self._rgba.g, self._rgba.b)  # noqa: E741
        if alpha is None:
            if self._rgba.alpha is None:
                return h, s, l
            else:
                return h, s, l, self._alpha_float()
        if alpha:
            return h, s, l, self._alpha_float()
        else:
            # alpha is False
            return h, s, l

    def _alpha_float(self) -> float:
        return 1 if self._rgba.alpha is None else self._rgba.alpha

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: Callable[[Any], CoreSchema]
    ) -> core_schema.CoreSchema:
        if hasattr(core_schema, "with_info_plain_validator_function"):
            return core_schema.with_info_plain_validator_function(
                cls._validate, serialization=core_schema.to_string_ser_schema()
            )
        return core_schema.general_plain_validator_function(
            cls._validate, serialization=core_schema.to_string_ser_schema()
        )

    @classmethod
    def _validate(cls, __input_value: Any, _: Any) -> Color:
        return cls(__input_value)

    def __str__(self) -> str:
        return self.as_named(fallback=True)

    def __repr_args__(self) -> Any:
        return [(None, self.as_named(fallback=True))] + [("rgb", self.as_rgb_tuple())]

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Color) and self.as_rgb_tuple() == other.as_rgb_tuple()

    def __hash__(self) -> int:
        return hash(self.as_rgb_tuple())


def parse_tuple(value: tuple[Any, ...]) -> RGBA:
    """Parse a tuple or list to get RGBA values.

    Args:
    ----
        value: A tuple or list.

    Returns:
    -------
        An `RGBA` tuple parsed from the input tuple.

    Raises:
    ------
        PydanticCustomError: If tuple is not valid.
    """
    if len(value) == 3:
        r, g, b = (parse_color_value(v) for v in value)
        return RGBA(r, g, b, None)
    elif len(value) == 4:
        r, g, b = (parse_color_value(v) for v in value[:3])
        return RGBA(r, g, b, parse_float_alpha(value[3]))
    else:
        raise PydanticCustomError(
            "color_error", "value is not a valid color: tuples must have length 3 or 4"
        )


def parse_str(value: str) -> RGBA:
    """
    Parse a string representing a color to an RGBA tuple.

    Possible formats for the input string include:

    * named color, see `COLORS_BY_NAME`
    * hex short eg. `<prefix>fff` (prefix can be `#`, `0x` or nothing)
    * hex long eg. `<prefix>ffffff` (prefix can be `#`, `0x` or nothing)
    * `rgb(<r>, <g>, <b>)`
    * `rgba(<r>, <g>, <b>, <a>)`
    * `transparent`

    Args:
    ----
        value: A string representing a color.

    Returns:
    -------
        An `RGBA` tuple parsed from the input string.

    Raises:
    ------
        ValueError: If the input string cannot be parsed to an RGBA tuple.
    """
    value_lower = value.lower()
    try:
        r, g, b = COLORS_BY_NAME[value_lower]
    except KeyError:
        pass
    else:
        return ints_to_rgba(r, g, b, None)

    m = re.fullmatch(r_hex_short, value_lower)
    if m:
        *rgb, a = m.groups()
        r, g, b = (int(v * 2, 16) for v in rgb)
        if a:
            alpha: float | None = int(a * 2, 16) / 255
        else:
            alpha = None
        return ints_to_rgba(r, g, b, alpha)

    m = re.fullmatch(r_hex_long, value_lower)
    if m:
        *rgb, a = m.groups()
        r, g, b = (int(v, 16) for v in rgb)
        if a:
            alpha = int(a, 16) / 255
        else:
            alpha = None
        return ints_to_rgba(r, g, b, alpha)

    m = re.fullmatch(r_rgb, value_lower) or re.fullmatch(r_rgb_v4_style, value_lower)
    if m:
        return ints_to_rgba(*m.groups())  # type: ignore

    m = re.fullmatch(r_hsl, value_lower) or re.fullmatch(r_hsl_v4_style, value_lower)
    if m:
        return parse_hsl(*m.groups())  # type: ignore

    if value_lower == "transparent":
        return RGBA(0, 0, 0, 0)

    raise PydanticCustomError(
        "color_error",
        "value is not a valid color: string not recognised as a valid color",
    )


def ints_to_rgba(
    r: int | str,
    g: int | str,
    b: int | str,
    alpha: float | None = None,
) -> RGBA:
    """
    Converts integer or string values for RGB color and an optional alpha value to an `RGBA` object.

    Args:
    ----
        r: An integer or string representing the red color value.
        g: An integer or string representing the green color value.
        b: An integer or string representing the blue color value.
        alpha: A float representing the alpha value. Defaults to None.

    Returns:
    -------
        An instance of the `RGBA` class with the corresponding color and alpha values.
    """
    return RGBA(
        parse_color_value(r),
        parse_color_value(g),
        parse_color_value(b),
        parse_float_alpha(alpha),
    )


def parse_color_value(value: int | str, max_val: int = 255) -> float:
    """
    Parse the color value provided and return a number between 0 and 1.

    Args:
    ----
        value: An integer or string color value.
        max_val: Maximum range value. Defaults to 255.

    Raises:
    ------
        PydanticCustomError: If the value is not a valid color.

    Returns:
    -------
        A number between 0 and 1.
    """
    try:
        color = float(value)
    except ValueError:
        raise PydanticCustomError(
            "color_error",
            "value is not a valid color: color values must be a valid number",
        )
    if 0 <= color <= max_val:
        return color / max_val
    else:
        raise PydanticCustomError(
            "color_error",
            "value is not a valid color: color values must be in the range 0 to {max_val}",
            {"max_val": max_val},
        )


def parse_float_alpha(value: None | str | float | int) -> float | None:
    """
    Parse an alpha value checking it's a valid float in the range 0 to 1.

    Args:
    ----
        value: The input value to parse.

    Returns:
    -------
        The parsed value as a float, or `None` if the value was None or equal 1.

    Raises:
    ------
        PydanticCustomError: If the input value cannot be successfully parsed as a float in the expected range.
    """
    if value is None:
        return None
    try:
        if isinstance(value, str) and value.endswith("%"):
            alpha = float(value[:-1]) / 100
        else:
            alpha = float(value)
    except ValueError:
        raise PydanticCustomError(
            "color_error",
            "value is not a valid color: alpha values must be a valid float",
        )

    if math.isclose(alpha, 1):
        return None
    elif 0 <= alpha <= 1:
        return alpha
    else:
        raise PydanticCustomError(
            "color_error",
            "value is not a valid color: alpha values must be in the range 0 to 1",
        )


def parse_hsl(
    h: str, h_units: str, sat: str, light: str, alpha: float | None = None
) -> RGBA:
    """
    Parse raw hue, saturation, lightness, and alpha values and convert to RGBA.

    Args:
    ----
        h: The hue value.
        h_units: The unit for hue value.
        sat: The saturation value.
        light: The lightness value.
        alpha: Alpha value.

    Returns:
    -------
        An instance of `RGBA`.
    """
    s_value, l_value = parse_color_value(sat, 100), parse_color_value(light, 100)

    h_value = float(h)
    if h_units in {None, "deg"}:
        h_value = h_value % 360 / 360
    elif h_units == "rad":
        h_value = h_value % rads / rads
    else:
        # turns
        h_value = h_value % 1

    r, g, b = hls_to_rgb(h_value, l_value, s_value)
    return RGBA(r, g, b, parse_float_alpha(alpha))


def float_to_255(c: float) -> int:
    """
    Converts a float value between 0 and 1 (inclusive) to an integer between 0 and 255 (inclusive).

    Args:
    ----
        c: The float value to be converted. Must be between 0 and 1 (inclusive).

    Returns:
    -------
        The integer equivalent of the given float value rounded to the nearest whole number.
    """
    return round(c * 255)


COLORS_BY_NAME = {
    "aliceblue": (240, 248, 255),
    "antiquewhite": (250, 235, 215),
    "aqua": (0, 255, 255),
    "aquamarine": (127, 255, 212),
    "azure": (240, 255, 255),
    "beige": (245, 245, 220),
    "bisque": (255, 228, 196),
    "black": (0, 0, 0),
    "blanchedalmond": (255, 235, 205),
    "blue": (0, 0, 255),
    "blueviolet": (138, 43, 226),
    "brown": (165, 42, 42),
    "burlywood": (222, 184, 135),
    "cadetblue": (95, 158, 160),
    "chartreuse": (127, 255, 0),
    "chocolate": (210, 105, 30),
    "coral": (255, 127, 80),
    "cornflowerblue": (100, 149, 237),
    "cornsilk": (255, 248, 220),
    "crimson": (220, 20, 60),
    "cyan": (0, 255, 255),
    "darkblue": (0, 0, 139),
    "darkcyan": (0, 139, 139),
    "darkgoldenrod": (184, 134, 11),
    "darkgray": (169, 169, 169),
    "darkgreen": (0, 100, 0),
    "darkgrey": (169, 169, 169),
    "darkkhaki": (189, 183, 107),
    "darkmagenta": (139, 0, 139),
    "darkolivegreen": (85, 107, 47),
    "darkorange": (255, 140, 0),
    "darkorchid": (153, 50, 204),
    "darkred": (139, 0, 0),
    "darksalmon": (233, 150, 122),
    "darkseagreen": (143, 188, 143),
    "darkslateblue": (72, 61, 139),
    "darkslategray": (47, 79, 79),
    "darkslategrey": (47, 79, 79),
    "darkturquoise": (0, 206, 209),
    "darkviolet": (148, 0, 211),
    "deeppink": (255, 20, 147),
    "deepskyblue": (0, 191, 255),
    "dimgray": (105, 105, 105),
    "dimgrey": (105, 105, 105),
    "dodgerblue": (30, 144, 255),
    "firebrick": (178, 34, 34),
    "floralwhite": (255, 250, 240),
    "forestgreen": (34, 139, 34),
    "fuchsia": (255, 0, 255),
    "gainsboro": (220, 220, 220),
    "ghostwhite": (248, 248, 255),
    "gold": (255, 215, 0),
    "goldenrod": (218, 165, 32),
    "gray": (128, 128, 128),
    "green": (0, 128, 0),
    "greenyellow": (173, 255, 47),
    "grey": (128, 128, 128),
    "honeydew": (240, 255, 240),
    "hotpink": (255, 105, 180),
    "indianred": (205, 92, 92),
    "indigo": (75, 0, 130),
    "ivory": (255, 255, 240),
    "khaki": (240, 230, 140),
    "lavender": (230, 230, 250),
    "lavenderblush": (255, 240, 245),
    "lawngreen": (124, 252, 0),
    "lemonchiffon": (255, 250, 205),
    "lightblue": (173, 216, 230),
    "lightcoral": (240, 128, 128),
    "lightcyan": (224, 255, 255),
    "lightgoldenrodyellow": (250, 250, 210),
    "lightgray": (211, 211, 211),
    "lightgreen": (144, 238, 144),
    "lightgrey": (211, 211, 211),
    "lightpink": (255, 182, 193),
    "lightsalmon": (255, 160, 122),
    "lightseagreen": (32, 178, 170),
    "lightskyblue": (135, 206, 250),
    "lightslategray": (119, 136, 153),
    "lightslategrey": (119, 136, 153),
    "lightsteelblue": (176, 196, 222),
    "lightyellow": (255, 255, 224),
    "lime": (0, 255, 0),
    "limegreen": (50, 205, 50),
    "linen": (250, 240, 230),
    "magenta": (255, 0, 255),
    "maroon": (128, 0, 0),
    "mediumaquamarine": (102, 205, 170),
    "mediumblue": (0, 0, 205),
    "mediumorchid": (186, 85, 211),
    "mediumpurple": (147, 112, 219),
    "mediumseagreen": (60, 179, 113),
    "mediumslateblue": (123, 104, 238),
    "mediumspringgreen": (0, 250, 154),
    "mediumturquoise": (72, 209, 204),
    "mediumvioletred": (199, 21, 133),
    "midnightblue": (25, 25, 112),
    "mintcream": (245, 255, 250),
    "mistyrose": (255, 228, 225),
    "moccasin": (255, 228, 181),
    "navajowhite": (255, 222, 173),
    "navy": (0, 0, 128),
    "oldlace": (253, 245, 230),
    "olive": (128, 128, 0),
    "olivedrab": (107, 142, 35),
    "orange": (255, 165, 0),
    "orangered": (255, 69, 0),
    "orchid": (218, 112, 214),
    "palegoldenrod": (238, 232, 170),
    "palegreen": (152, 251, 152),
    "paleturquoise": (175, 238, 238),
    "palevioletred": (219, 112, 147),
    "papayawhip": (255, 239, 213),
    "peachpuff": (255, 218, 185),
    "peru": (205, 133, 63),
    "pink": (255, 192, 203),
    "plum": (221, 160, 221),
    "powderblue": (176, 224, 230),
    "purple": (128, 0, 128),
    "red": (255, 0, 0),
    "rosybrown": (188, 143, 143),
    "royalblue": (65, 105, 225),
    "saddlebrown": (139, 69, 19),
    "salmon": (250, 128, 114),
    "sandybrown": (244, 164, 96),
    "seagreen": (46, 139, 87),
    "seashell": (255, 245, 238),
    "sienna": (160, 82, 45),
    "silver": (192, 192, 192),
    "skyblue": (135, 206, 235),
    "slateblue": (106, 90, 205),
    "slategray": (112, 128, 144),
    "slategrey": (112, 128, 144),
    "snow": (255, 250, 250),
    "springgreen": (0, 255, 127),
    "steelblue": (70, 130, 180),
    "tan": (210, 180, 140),
    "teal": (0, 128, 128),
    "thistle": (216, 191, 216),
    "tomato": (255, 99, 71),
    "turquoise": (64, 224, 208),
    "violet": (238, 130, 238),
    "wheat": (245, 222, 179),
    "white": (255, 255, 255),
    "whitesmoke": (245, 245, 245),
    "yellow": (255, 255, 0),
    "yellowgreen": (154, 205, 50),
}

COLORS_BY_VALUE = {v: k for k, v in COLORS_BY_NAME.items()}



from __future__ import annotations

import os
import re
import subprocess
import sys
from pathlib import Path
from shutil import rmtree
from typing import Any

from xsdata.codegen.writer import CodeWriter
from xsdata.models import config as cfg
from xsdata.utils import text

from ome_autogen import _util
from ome_autogen._util import camel_to_snake
from ome_autogen.generator import OmeGenerator
from ome_autogen.overrides import MIXINS
from ome_autogen.transformer import OMETransformer
from xsdata_pydantic_basemodel.config import GeneratorOutput

# these are normally "reserved" names that we want to allow as field names
ALLOW_RESERVED_NAMES = {"type", "Type", "Union"}
# format key used to register our custom OmeGenerator
OME_FORMAT = "OME"

PYDANTIC_SUPPORT = os.getenv("PYDANTIC_SUPPORT", "both")
RUFF_LINE_LENGTH = 88
RUFF_TARGET_VERSION = "py38"
OUTPUT_PACKAGE = "ome_types._autogenerated.ome_2016_06"
DO_MYPY = os.environ.get("OME_AUTOGEN_MYPY", "0") == "1" or "--mypy" in sys.argv
SRC_PATH = Path(__file__).parent.parent
SCHEMA_FILE = (SRC_PATH / "ome_types" / "ome-2016-06.xsd").absolute()
RUFF_IGNORE: list[str] = [
    "D101",  # Missing docstring in public class
    "D106",  # Missing docstring in public nested class
    "D205",  # 1 blank line required between summary line and description
    "D404",  # First word of the docstring should not be This
    "E501",  # Line too long
    "S105",  # Possible hardcoded password
    "RUF002",  # ambiguous-unicode-character-docstring
]


def get_config(
    package: str, kw_only: bool = True, compound_fields: bool = False
) -> cfg.GeneratorConfig:
    """Return a GeneratorConfig for the OME schema."""
    # ALLOW "type" to be used as a field name
    text.stop_words.difference_update(ALLOW_RESERVED_NAMES)

    # use our own camel_to_snake
    # Our's interprets adjacent capital letters as two words
    # NameCase.SNAKE: 'PositionXUnit' -> 'position_xunit'
    # camel_to_snake: 'PositionXUnit' -> 'position_x_unit'
    cfg.__name_case_func__["snakeCase"] = camel_to_snake

    #  critical to be able to use the format="OME"
    CodeWriter.register_generator(OME_FORMAT, OmeGenerator)

    mixins = []
    for class_name, import_string, prepend in MIXINS:
        mixins.append(
            cfg.GeneratorExtension(
                type=cfg.ExtensionType.CLASS,
                class_name=class_name,
                import_string=import_string,
                prepend=prepend,
            )
        )

    keep_case = cfg.NameConvention(cfg.NameCase.ORIGINAL, "type")
    return cfg.GeneratorConfig(
        output=GeneratorOutput(
            package=package,
            # format.value lets us use our own generator
            # kw_only is important, it makes required fields actually be required
            format=cfg.OutputFormat(value=OME_FORMAT, kw_only=kw_only),
            structure_style=cfg.StructureStyle.CLUSTERS,
            docstring_style=cfg.DocstringStyle.NUMPY,
            compound_fields=cfg.CompoundFields(enabled=compound_fields),
            # whether to create models that work for both pydantic 1 and 2
            pydantic_support=PYDANTIC_SUPPORT,  # type: ignore
        ),
        # Add our mixins
        extensions=cfg.GeneratorExtensions(mixins),
        # Don't convert things like XMLAnnotation to XmlAnnotation
        conventions=cfg.GeneratorConventions(class_name=keep_case),
    )


def build_model(
    output_dir: Path | str = SRC_PATH,
    schema_file: Path | str = SCHEMA_FILE,
    target_package: str = OUTPUT_PACKAGE,
    ruff_ignore: list[str] = RUFF_IGNORE,
    do_formatting: bool = True,
    do_mypy: bool = DO_MYPY,
) -> None:
    """Convert the OME schema to a python model."""
    config = get_config(target_package)
    transformer = OMETransformer(print=False, config=config)

    _print_gray(f"Processing {getattr(schema_file ,'name', schema_file)}...")
    transformer.process_sources([Path(schema_file).resolve().as_uri()])

    package_dir = str(Path(output_dir) / OUTPUT_PACKAGE.replace(".", "/"))
    rmtree(package_dir, ignore_errors=True)
    with _util.cd(output_dir):  # xsdata doesn't support output path
        _print_gray("Writing Files...")
        transformer.process_classes()

    _build_typed_dicts(package_dir)
    if do_formatting:
        _fix_formatting(package_dir, ruff_ignore)

    if do_mypy:
        _check_mypy(package_dir)

    _print_green(f"OME python model created at {OUTPUT_PACKAGE}")


def _fix_formatting(package_dir: str, ruff_ignore: list[str] = RUFF_IGNORE) -> None:
    _print_gray("Running ruff check...")

    ruff_chk = ["ruff", "check", "-q", "--fix", "--unsafe-fixes", package_dir]
    ruff_chk.extend(f"--ignore={ignore}" for ignore in ruff_ignore)
    subprocess.check_call(ruff_chk)  # noqa S

    _print_gray("Running ruff format...")
    ruff_fmt = [
        "ruff",
        "format",
        "-q",
        f"--line-length={RUFF_LINE_LENGTH}",
        f"--target-version={RUFF_TARGET_VERSION}",
    ]
    ruff_fmt.extend([str(x) for x in Path(package_dir).rglob("*.py")])
    subprocess.check_call(ruff_fmt)  # noqa S


def _check_mypy(package_dir: str) -> None:
    _print_gray("Running mypy ...")

    # FIXME: the call-overload disable is due to Field() in pydantic/pydantic-compat.
    mypy = ["mypy", package_dir, "--strict", "--disable-error-code", "call-overload"]
    try:
        subprocess.check_output(mypy, stderr=subprocess.STDOUT)  # noqa S
    except subprocess.CalledProcessError as e:  # pragma: no cover
        raise RuntimeError(f"mypy errors:\n\n{e.output.decode()}") from e


def _print_gray(text: str) -> None:
    if os.name != "nt":
        # UnicodeEncodeError: 'charmap' codec can't encode character '\u2713'
        text = f"\033[90m\033[1m{text}\033[0m"
    print(text)


def _print_green(text: str) -> None:
    if os.name != "nt":
        # UnicodeEncodeError: 'charmap' codec can't encode character '\u2713'
        text = f"\033[92m\033[1m{text}\033[0m"
    print(text)


KWARGS_MODULE = """
from __future__ import annotations
from typing_extensions import TypeAlias
from typing import Union, List, TypedDict
from datetime import datetime
import ome_types.model as ome

class RefDict(TypedDict):
    id: str
"""


def _build_typed_dicts(package_dir: str) -> None:
    """Create a TypedDict class for each OMEType subclass.

    Useful for passing kwargs to the constructors.

    https://peps.python.org/pep-0692/

    def foo(**kwargs: Unpack[ome.ImageDict]) -> None:
        ...
    """
    # sourcery skip: assign-if-exp, reintroduce-else
    try:
        from pydantic._internal._repr import display_as_type
    except ImportError:
        # don't try to do this on pydantic1
        return
    if PYDANTIC_SUPPORT == "v1":
        return

    from ome_types import model
    from ome_types._mixins._base_type import OMEType

    ome_models = {
        name: obj
        for name, obj in vars(model).items()
        if isinstance(obj, type) and issubclass(obj, OMEType) and obj.__annotations__
    }

    def _disp_type(obj: Any) -> str:
        x = display_as_type(obj).replace("NoneType", "None")
        if "ForwardRef" in x:
            #  replace "List[ForwardRef('Map.M')]" with "List[Map.M]"
            x = re.sub(r"ForwardRef\('([a-zA-Z_.]*)'\)", r"\1", x)
        x = re.sub(r"ome_types\._autogenerated\.ome_2016_06.[^.]+.", "", x)
        return x

    # add TypedDicts for all models
    module = KWARGS_MODULE
    SUFFIX = "Dict"
    CLASS = "class {name}(TypedDict, total=False):\n    {fields}\n\n"
    for cls_name, m in sorted(ome_models.items()):
        if cls_name.endswith("Ref"):
            module += f"{cls_name}: TypeAlias = RefDict\n"
        else:
            _fields = [
                f"{k}: {_disp_type(v.annotation)}"
                # this type ignore indicates something that may break in pydantic 3
                # but for now, it's confusing and I think it's an error
                for k, v in sorted(m.model_fields.items())  # type: ignore
            ]
            if _fields:
                module += CLASS.format(
                    name=f"{m.__name__}{SUFFIX}", fields="\n    ".join(_fields)
                )
            else:
                module += (
                    f"class {m.__name__}{SUFFIX}(TypedDict, total=False):\n    pass\n\n"
                )

    # fix name spaces
    # prefix all remaining capitalized words with ome.
    def _repl(match: re.Match) -> str:
        word = match[1]
        if word in {"None", "True", "False", "Union", "List", "TypedDict", "TypeAlias"}:
            return word
        if word.endswith(SUFFIX):
            return word
        if word in ome_models:
            return word + SUFFIX
        # the rest are enums, they can be passed as strings
        return f"ome.{word} | str"

    module = re.sub(r"\b([A-Z][a-zA-Z_^.]*)\b", _repl, module)
    (Path(package_dir) / "kwargs.py").write_text(module)



from typing import TYPE_CHECKING, Any, Dict, Iterator, List, MutableMapping, Optional

try:
    from pydantic import model_serializer
except ImportError:
    model_serializer = None  # type: ignore


if TYPE_CHECKING:
    from typing import Protocol

    from ome_types._autogenerated.ome_2016_06.map import Map

    class HasMsProtocol(Protocol):
        @property
        def ms(self) -> List["Map.M"]: ...


class MapMixin(MutableMapping[str, Optional[str]]):
    def __delitem__(self: "HasMsProtocol", key: str) -> None:
        for m in self.ms:
            if m.k == key:
                self.ms.remove(m)
                return

    def __len__(self: "HasMsProtocol") -> int:
        return len(self.ms)

    def __iter__(self: "HasMsProtocol") -> Iterator[str]:
        yield from (m.k for m in self.ms if m.k is not None)

    def __getitem__(self: "HasMsProtocol", key: str) -> Optional[str]:
        return next((m.value for m in self.ms if m.k == key), None)

    def __setitem__(self: "HasMsProtocol", key: str, value: Optional[str]) -> None:
        for m in self.ms:
            if m.k == key:
                m.value = value or ""
                return
        from ome_types.model import Map

        self.ms.append(Map.M(k=key, value=value))

    def _pydict(self: "HasMsProtocol", **kwargs: Any) -> Dict[str, str]:
        return {m.k: m.value for m in self.ms if m.k is not None}

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        return self._pydict()  # type: ignore

    if model_serializer is not None:

        @model_serializer(mode="wrap")
        def serialize_root(self, handler, _info) -> dict:  # type: ignore
            return self._pydict()  # type: ignore



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

# from ome_types._mixins._base_type import AUTO_SEQUENCE
# avoiding import to avoid build-time dependency on the ome-types package
AUTO_SEQUENCE = "__auto_sequence__"

MIXIN_MODULE = "ome_types._mixins"
# class_name, import_string, whether-to-prepend or append to the existing Bases
MIXINS: list[tuple[str, str, bool]] = [
    (".*", f"{MIXIN_MODULE}._base_type.OMEType", False),  # base type on every class
    ("OME", f"{MIXIN_MODULE}._ome.OMEMixin", True),
    ("Instrument", f"{MIXIN_MODULE}._instrument.InstrumentMixin", False),
    ("Reference", f"{MIXIN_MODULE}._reference.ReferenceMixin", True),
    ("Map", f"{MIXIN_MODULE}._map_mixin.MapMixin", False),
    ("Union", f"{MIXIN_MODULE}._collections.ShapeUnionMixin", True),
    (
        "StructuredAnnotations",
        f"{MIXIN_MODULE}._collections.StructuredAnnotationsMixin",
        True,
    ),
    ("(Shape|ManufacturerSpec|Annotation)", f"{MIXIN_MODULE}._kinded.KindMixin", True),
]


@dataclass
class Ovr:
    """Override settings for a given name in the XSD."""

    # Name of an attribute or element in the XSD
    # (this is autopopulated by the key in the OVERRIDES dict, so as to ensure
    # that we don't have more than one override for a given name in the XSD.)
    name: str = ""
    # additional lines to add to generated classes
    # Note: imports for these methods must be added to the IMPORT_PATTERNS elswhere.
    add_lines: Sequence[str] = ()
    # when name is found as a type in the XSD, the type will be overridden with this
    # e.g. "Color" -> ("ome_types.model._color", "Color")
    type_override: tuple[str, str] | None = None
    # When name is used as a type hint, it should never be Optional[]
    never_optional: bool = False
    # default value for a field with this type
    default: str | None = None
    # default factory for a field with this type
    default_factory: str = ""


OVERRIDES: dict[str, Ovr] = {
    "ID": Ovr(default=repr(AUTO_SEQUENCE)),
    "Union": Ovr(never_optional=True, default_factory="lambda: ROI.Union()"),
    "StructuredAnnotations": Ovr(
        never_optional=True, default_factory="StructuredAnnotations"
    ),
    "FillColor": Ovr(type_override=("ome_types.model._color", "Color")),
    "StrokeColor": Ovr(
        type_override=("ome_types.model._color", "Color"),
    ),
    "Color": Ovr(
        type_override=("ome_types.model._color", "Color"),
        default_factory="Color",
    ),
    "BinData": Ovr(
        add_lines=[
            "_vbindata = model_validator(mode='before')(bin_data_root_validator)"
        ],
    ),
    "Value": Ovr(
        add_lines=["_vany = field_validator('any_elements')(any_elements_validator)"],
    ),
    "Pixels": Ovr(
        add_lines=["_vpix = model_validator(mode='before')(pixels_root_validator)"],
    ),
    "XMLAnnotation": Ovr(
        add_lines=[
            "_vval = field_validator('value', mode='before')(xml_value_validator)"
        ],
    ),
    "PixelType": Ovr(add_lines=["numpy_dtype = property(pixel_type_to_numpy_dtype)"]),
    "OME": Ovr(
        add_lines=[
            "_v_structured_annotations = field_validator('structured_annotations', mode='before')(validate_structured_annotations)"
        ],
    ),
    "ROI": Ovr(
        add_lines=[
            "_v_shape_union = field_validator('union', mode='before')(validate_shape_union)"
        ],
    ),
    "Map": Ovr(
        add_lines=[
            "_v_map = model_validator(mode='before')(validate_map_annotation)",
            "dict: ClassVar = MapMixin._pydict",
            "__iter__: ClassVar = MapMixin.__iter__",
        ],
    ),
}

# autopopulate the name attribute
for k, v in OVERRIDES.items():
    v.name = k
    if v.never_optional and not (v.default or v.default_factory):
        raise ValueError(
            f"never_optional overrides must have a default or default_factory: {k}"
        )

# Quick lookup mapping XSD name to overridden type name
# e.g. {'FillColor': 'Color', 'StrokeColor': 'Color', 'Color': 'Color'}
OVERRIDE_ELEM_TO_CLASS: dict[str, str] = {}

# set of element names that should never be optional
# (but always have default_factories)
NEVER_OPTIONAL = {x for x in OVERRIDES if OVERRIDES[x].never_optional}

# import patterns
# key is a module path.  Value is a map of type names to a list of patterns to match
# in the generated code.  If a pattern is found, the type will be imported from the
# module specified, in the file where the pattern was found.
IMPORT_PATTERNS: dict[str, dict[str, list[str]]] = {
    "typing": {"ClassVar": [": ClassVar"]},
    "ome_types._mixins._util": {"new_uuid": ["default_factory=new_uuid"]},
    "datetime": {"datetime": ["datetime"]},
    "pydantic": {"validator": ["validator("]},
    "pydantic_compat": {
        "model_validator": ["model_validator("],
        "field_validator": ["field_validator("],
    },
    "ome_types._mixins._validators": {
        "any_elements_validator": ["any_elements_validator"],
        "bin_data_root_validator": ["bin_data_root_validator"],
        "pixel_type_to_numpy_dtype": ["pixel_type_to_numpy_dtype"],
        "pixels_root_validator": ["pixels_root_validator"],
        "validate_map_annotation": ["validate_map_annotation"],
        "validate_shape_union": ["validate_shape_union"],
        "validate_structured_annotations": ["validate_structured_annotations"],
        "xml_value_validator": ["xml_value_validator"],
    },
}


for override in OVERRIDES.values():
    if override.type_override:
        module, type_name = override.type_override
        mod_entry = IMPORT_PATTERNS.setdefault(module, {})
        mod_entry[type_name] = [f": {type_name} =", f": Optional[{type_name}] ="]
        OVERRIDE_ELEM_TO_CLASS[override.name] = type_name


def get(name: str, default: Ovr | None = None) -> Ovr | None:
    """Get an override by name.  If not found, return default."""
    return OVERRIDES.get(name, default)
